import numpy as np

from ._param_grid import ParamGrid
from ._parameter_search import ParameterSearch
from ._stagnation_stopping_task import StagnationStoppingTask

from niapy.algorithms.basic import FireflyAlgorithm, BatAlgorithm, GreyWolfOptimizer
from niapy.algorithms.modified import HybridBatAlgorithm, HybridSelfAdaptiveBatAlgorithm

from sklearn.model_selection._search import BaseSearchCV, check_cv, is_classifier

SUPPORTED_ALGORITHMS = ['fa', 'gwo', 'ba', 'hba', 'hsaba']


class NatureInspiredSearchCV(BaseSearchCV):
    def fit(self, X, y=None, *, groups=None, **fit_params):
        self.cv_orig = check_cv(self.cv, y, classifier=is_classifier(self.estimator))
        self.n_splits = self.cv_orig.get_n_splits(X, y, groups)
        return super().fit(X, y, groups=groups, **fit_params)

    def _run_search(self, evaluate_candidates):
        self.__param_grid = ParamGrid(self.param_grid)
        self.__algorithm = self.__get_algorithm(self.algorithm, self.population_size, self.random_state)
        self.__print_run_search_log()
        problem = ParameterSearch(evaluate_candidates, self.__param_grid)

        self.optimization_logs_ = []

        for i in range(1, self.runs + 1):
            def iteration_finished_callback(iteration, stagnation_gen_count):
                stagnation = f' | Stagnation {stagnation_gen_count}'

                if self.max_stagnating_gen > 1:
                    stagnation = f'{stagnation}/{self.max_stagnating_gen}'

                if self.verbose > 1:
                    print(f'Run {i}/{self.runs}, Iteration {iteration}/{self.max_n_gen}{stagnation}'
                          f' | There were {problem.evaluation_count} candidates trained until now')

            task = StagnationStoppingTask(
                max_iters=self.max_n_gen,
                max_stagnating_gen=self.max_stagnating_gen,
                problem=problem,
                iteration_finished_callback=iteration_finished_callback
            )

            self.__algorithm.run(task=task)

            if self.__algorithm.exception:
                raise self.__algorithm.exception

            # invert scores in the optimization logs
            self.optimization_logs_.append([(log[0], log[1] * -1) for log in task.optimization_logs_])

            if self.verbose > 0:
                print(f'Run {i}/{self.runs} finished')

        candidates = problem.evaluation_count

        if self.cv is None:
            print(f'Optimization finished, {candidates} candidates were fitted')
        else:
            print(f'Optimization finished, {candidates} candidates were fitted '
                  f'(totalling {candidates * self.n_splits} fits)')

    def __init__(self, estimator, param_grid, algorithm='hba', population_size=50, max_n_gen=100, runs=3,
                 max_stagnating_gen=20, scoring=None, n_jobs=None, refit=True, cv=None, verbose=0,
                 random_state=None,
                 pre_dispatch='2*n_jobs', error_score=np.nan, return_train_score=True):
        super().__init__(estimator, scoring=scoring, n_jobs=n_jobs, refit=refit, cv=cv, pre_dispatch=pre_dispatch,
                         error_score=error_score, return_train_score=return_train_score)

        self.algorithm = algorithm
        self.cv = cv
        self.max_n_gen = max_n_gen
        self.max_stagnating_gen = max_stagnating_gen
        self.param_grid = param_grid
        self.population_size = population_size
        self.random_state = random_state
        self.runs = runs
        self.verbose = verbose
        self.optimization_logs_ = None
        self.cv_orig = None
        self.n_splits = None

    def __print_run_search_log(self):
        candidates = self.__param_grid.get_number_of_candidates()

        if self.cv is not None:
            print(f'Fitting {self.cv} folds for some of the {candidates} candidates, '
                  f'which might total in {candidates * self.n_splits} fits')
        else:
            print(f'Fitting at most {candidates} candidates')

    @staticmethod
    def __get_algorithm(algorithm, population_size, random_state):
        if isinstance(algorithm, str):
            if algorithm not in SUPPORTED_ALGORITHMS:
                raise ValueError(f'"{algorithm}" is not in supported algorithms: {", ".join(SUPPORTED_ALGORITHMS)}')

            algorithm_obj = None

            if algorithm == 'fa':
                algorithm_obj = FireflyAlgorithm(seed=random_state)
                algorithm_obj.set_parameters(alpha=1, betamin=1, gamma=2)
            elif algorithm == 'ba':
                algorithm_obj = BatAlgorithm(seed=random_state)
                algorithm_obj.set_parameters(A=0.9, r=0.1, Qmin=0.0, Qmax=2.0)
            elif algorithm == 'hba':
                algorithm_obj = HybridBatAlgorithm(seed=random_state)
                algorithm_obj.set_parameters(A=0.9, r=0.1, Qmin=0.0, Qmax=2.0)
            elif algorithm == 'hsaba':
                algorithm_obj = HybridSelfAdaptiveBatAlgorithm(seed=random_state)
                algorithm_obj.set_parameters(A=0.9, r=0.1, Qmin=0.0, Qmax=2.0)
            elif algorithm == 'gwo':
                algorithm_obj = GreyWolfOptimizer(seed=random_state)

            algorithm_obj.set_parameters(NP=population_size)

            return algorithm_obj

        return algorithm



from niapy.task import Task, OptimizationType


class StagnationStoppingTask(Task):
    prev_x_f = None

    def __init__(self, max_stagnating_gen=None, iteration_finished_callback=None, **kwargs):
        super().__init__(**kwargs,
                         optimization_type=OptimizationType.MINIMIZATION)

        if not isinstance(max_stagnating_gen, int):
            if max_stagnating_gen is not None:
                raise ValueError('max_stagnating_gen should be int or None!')

        if max_stagnating_gen is not None and max_stagnating_gen < 1:
            raise ValueError('max_stagnating_gen greater or equal to 1!')

        self.max_stagnating_gen = max_stagnating_gen

        # logs array of (generation, score)
        self.optimization_logs_ = []

        self.stagnating_gen_count = 0
        self.stagnating_gen_eval_count = 0
        self.is_stagnating = False
        self.iteration_finished_callback = iteration_finished_callback

    def eval_stagnation(self):
        # the better score is "lower" since the optimization type is "MINIMIZATION"
        if self.prev_x_f is None or self.x_f < self.prev_x_f:
            self.prev_x_f = self.x_f
            self.stagnating_gen_count = 0
        else:
            self.stagnating_gen_count += 1

        if self.stagnating_gen_count >= self.max_stagnating_gen:
            self.is_stagnating = True

    def eval(self, A):
        x_f = super().eval(A)
        self.optimization_logs_.append((self.iters, x_f))
        return x_f

    def stopping_condition(self):
        if self.is_stagnating:
            return True

        return super().stopping_condition()

    def next_iter(self):
        super().next_iter()

        if self.max_stagnating_gen:
            self.eval_stagnation()

            if self.iteration_finished_callback:
                self.iteration_finished_callback(self.iters, self.stagnating_gen_count)



import seaborn as sns
import pandas as pd

import matplotlib.pyplot as plt


def score_by_generation_violinplot(nia_search, run=0, ax=None, ylim=None):
    if run >= len(nia_search.optimization_logs_):
        raise ValueError(f'Specified run ({run}) does not exist!')

    df = pd.DataFrame(nia_search.optimization_logs_[run], columns=['generation', 'score'])

    if ax is None:
        _, ax = plt.subplots()

    if ylim is not None:
        ax.set(ylim=ylim)

    return sns.violinplot(x="generation", y="score", data=df)



import seaborn as sns
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt

ALLOWED_METRICS = ['min', 'max', 'median', 'mean']


def score_by_generation_lineplot(nia_search, metric='max', ax=None, ylim=None):
    if metric not in ALLOWED_METRICS:
        raise ValueError(f'{metric} is not in the allowed metrics! ({ALLOWED_METRICS})')

    optimization_logs_by_runs = []

    for run, optimization_logs in enumerate(nia_search.optimization_logs_):
        for log in optimization_logs:
            optimization_logs_by_runs.append((run, *log))

    df = pd.DataFrame(optimization_logs_by_runs, columns=['run', 'generation', 'score'])

    score_df = df.groupby(['run', 'generation']) \
        .agg({'score': [np.min, np.max, np.mean, np.median]}) \
        .reset_index()

    # remove multilevel columns
    score_df.columns = [[col_ for col_ in col if len(col_)][-1] for col in score_df.columns.values]

    # rename columns with weird names
    score_df = score_df.rename(columns={'amax': 'max', 'amin': 'min'})
    # rename score column
    score_df = score_df.rename(columns={metric: f'{metric} score'})

    if ax is None:
        _, ax = plt.subplots()

    if ylim is not None:
        ax.set(ylim=ylim)

    return sns.lineplot(x="generation", y=f'{metric} score', data=score_df, hue='run')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sklearn_nature_inspired_algorithms',
 'sklearn_nature_inspired_algorithms.helpers',
 'sklearn_nature_inspired_algorithms.model_selection']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5,<4.0.0',
 'niapy==2.0.1',
 'numpy>=1.22,<2.0.0',
 'pandas>=1.4,<2.0.0',
 'scikit-learn>=1.0.2',
 'scipy>=1.7.3',
 'seaborn>=0.11,<1.0.0',
 'toml>=0.10,<1.0.0']

setup_kwargs = {
    'name': 'sklearn-nature-inspired-algorithms',
    'version': '0.8.2',
    'description': 'Search using nature inspired algorithms over specified parameter values for an sklearn estimator.',
    'long_description': '# Nature-Inspired Algorithms for scikit-learn\n\n[![CI](https://github.com/timzatko/Sklearn-Nature-Inspired-Algorithms/workflows/CI/badge.svg?branch=master)](https://github.com/timzatko/Sklearn-Nature-Inspired-Algorithms/actions?query=workflow:CI+branch:master)\n[![Maintainability](https://api.codeclimate.com/v1/badges/ed99e5c765bf5c95d716/maintainability)](https://codeclimate.com/github/timzatko/Sklearn-Nature-Inspired-Algorithms/maintainability)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/sklearn-nature-inspired-algorithms)\n[![PyPI version](https://badge.fury.io/py/sklearn-nature-inspired-algorithms.svg)](https://pypi.org/project/sklearn-nature-inspired-algorithms/)\n[![PyPI downloads](https://img.shields.io/pypi/dm/sklearn-nature-inspired-algorithms)](https://pypi.org/project/sklearn-nature-inspired-algorithms/)\n[![Fedora package](https://img.shields.io/fedora/v/python3-sklearn-nature-inspired-algorithms?color=blue&label=Fedora%20Linux&logo=fedora)](https://src.fedoraproject.org/rpms/python-sklearn-nature-inspired-algorithms)\n\nNature-inspired algorithms for hyper-parameter tuning of [scikit-learn](https://github.com/scikit-learn/scikit-learn) models. This package uses algorithms implementation from [NiaPy](https://github.com/NiaOrg/NiaPy). \n\n## Installation\n\n```shell script\n$ pip install sklearn-nature-inspired-algorithms\n```\n\nTo install this package on Fedora, run:\n\n```sh\n$ dnf install python3-sklearn-nature-inspired-algorithms\n```\n\n## Usage\n\nThe usage is similar to using sklearn\'s `GridSearchCV`. Refer to the [documentation](https://sklearn-nature-inspired-algorithms.readthedocs.io/en/stable/) for more detailed guides and more examples.\n\n```python\nfrom sklearn_nature_inspired_algorithms.model_selection import NatureInspiredSearchCV\nfrom sklearn.ensemble import RandomForestClassifier\n\nparam_grid = { \n    \'n_estimators\': range(20, 100, 20), \n    \'max_depth\': range(2, 40, 2),\n    \'min_samples_split\': range(2, 20, 2), \n    \'max_features\': ["auto", "sqrt", "log2"],\n}\n\nclf = RandomForestClassifier(random_state=42)\n\nnia_search = NatureInspiredSearchCV(\n    clf,\n    param_grid,\n    algorithm=\'hba\', # hybrid bat algorithm\n    population_size=50,\n    max_n_gen=100,\n    max_stagnating_gen=10,\n    runs=3,\n    random_state=None, # or any number if you want same results on each run\n)\n\nnia_search.fit(X_train, y_train)\n\n# the best params are stored in nia_search.best_params_\n# finally you can train your model with best params from nia search\nnew_clf = RandomForestClassifier(**nia_search.best_params_, random_state=42)\n```\n\nAlso you plot the search process with _line plot_ or _violin plot_.\n\n```python\nfrom sklearn_nature_inspired_algorithms.helpers import score_by_generation_lineplot, score_by_generation_violinplot\n\n# line plot will plot all of the runs, you can specify the metric to be plotted (\'min\', \'max\', \'median\', \'mean\')\nscore_by_generation_lineplot(nia_search, metric=\'max\')\n\n# in violin plot you need to specify the run to be plotted\nscore_by_generation_violinplot(nia_search, run=0)\n```\n\nJupyter notebooks with full examples are available in [here](examples/notebooks).\n\n### Using a Custom Nature-Inspired Algorithm\n\nIf you do not want to use any of the pre-defined algorithm configurations, you can use any algorithm from the  [NiaPy](https://github.com/NiaOrg/NiaPy) collection.\nThis will allow you to have more control of the algorithm behavior. \nRefer to their [documentation](https://niapy.readthedocs.io/en/latest/) and [examples](https://github.com/NiaOrg/NiaPy/tree/master/examples) for the usage. \n\n__Note:__ Use version >2.x.x of NiaPy package\n\n```python\nfrom niapy.algorithms.basic import GeneticAlgorithm\n\nalgorithm = GeneticAlgorithm() # when custom algorithm is provided random_state is ignored\nalgorithm.set_parameters(NP=50, Ts=5, Mr=0.25)\n\nnia_search = NatureInspiredSearchCV(\n    clf,\n    param_grid,\n    algorithm=algorithm,\n    population_size=50,\n    max_n_gen=100,\n    max_stagnating_gen=20,\n    runs=3,\n)\n\nnia_search.fit(X_train, y_train)\n```\n\n## Contributing \n\nDetailed information on the contribution guidelines are in the [CONTRIBUTING.md](./CONTRIBUTING.md).\n',
    'author': 'Timotej Zatko',
    'author_email': 'timi.zatko@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/timzatko/Sklearn-Nature-Inspired-Algorithms',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<3.11',
}


setup(**setup_kwargs)
